@@ -1,36 +1,34 @@
-# -*- coding: utf-8 -*-
 """Command line tool for managing tmux workspaces and tmuxp configurations.
 
 tmuxp.cli
 ~~~~~~~~~
 
 """
-from __future__ import absolute_import
-
+import importlib
 import logging
 import os
 import platform
 import sys
+from subprocess import call
 
 import click
 import kaptan
 from click.exceptions import FileError
 
+from libtmux import __version__ as libtmux_version
 from libtmux.common import (
+    get_version,
     has_gte_version,
     has_minimum_version,
-    which,
-    get_version,
     tmux_cmd,
+    which,
 )
 from libtmux.exc import TmuxCommandNotFound
 from libtmux.server import Server
 
-from libtmux import __version__ as libtmux_version
-
-from . import config, exc, log, util, __file__ as tmuxp_path
+from . import __file__ as tmuxp_path, config, exc, log, util
 from .__about__ import __version__
-from ._compat import PY3, PYMINOR, string_types
+from ._compat import PY3, PYMINOR
 from .workspacebuilder import WorkspaceBuilder, freeze
 
 logger = logging.getLogger(__name__)
@@ -262,7 +260,7 @@ def scan_config_argument(ctx, param, value, config_dir=None):
         tmuxp_echo(ctx.get_help(), color=ctx.color)
         ctx.exit()
 
-    if isinstance(value, string_types):
+    if isinstance(value, str):
         value = scan_config(value, config_dir=config_dir)
 
     elif isinstance(value, tuple):
@@ -396,13 +394,48 @@ def scan_config(config, config_dir=None):
     return config
 
 
-def _reattach(session):
+def load_plugins(sconf):
+    """
+    Load and return plugins in config
+    """
+    plugins = []
+    if 'plugins' in sconf:
+        for plugin in sconf['plugins']:
+            try:
+                module_name = plugin.split('.')
+                module_name = '.'.join(module_name[:-1])
+                plugin_name = plugin.split('.')[-1]
+                plugin = getattr(importlib.import_module(module_name), plugin_name)
+                plugins.append(plugin())
+            except exc.TmuxpPluginException as error:
+                if not click.confirm(
+                    '%sSkip loading %s?'
+                    % (click.style(str(error), fg='yellow'), plugin_name),
+                    default=True,
+                ):
+                    click.echo(
+                        click.style('[Not Skipping] ', fg='yellow')
+                        + 'Plugin versions constraint not met. Exiting...'
+                    )
+                    sys.exit(1)
+            except Exception as error:
+                click.echo(
+                    click.style('[Plugin Error] ', fg='red')
+                    + "Couldn\'t load {0}\n".format(plugin)
+                    + click.style('{0}'.format(error), fg='yellow')
+                )
+                sys.exit(1)
+
+    return plugins
+
+
+def _reattach(builder):
     """
     Reattach session (depending on env being inside tmux already or not)
 
     Parameters
     ----------
-    session : :class:`libtmux.Session`
+    builder: :class:`workspacebuilder.WorkspaceBuilder`
 
     Notes
     -----
@@ -412,11 +445,96 @@ def _reattach(session):
 
     If not, ``tmux attach-session`` loads the client to the target session.
     """
+    for plugin in builder.plugins:
+        plugin.reattach(builder.session)
+        proc = builder.session.cmd('display-message', '-p', "'#S'")
+        for line in proc.stdout:
+            print(line)
+
     if 'TMUX' in os.environ:
-        session.switch_client()
+        builder.session.switch_client()
 
     else:
-        session.attach_session()
+        builder.session.attach_session()
+
+
+def _load_attached(builder, detached):
+    """
+    Load config in new session
+
+    Parameters
+    ----------
+    builder: :class:`workspacebuilder.WorkspaceBuilder`
+    detached : bool
+    """
+    builder.build()
+
+    if 'TMUX' in os.environ:  # tmuxp ran from inside tmux
+        # unset TMUX, save it, e.g. '/tmp/tmux-1000/default,30668,0'
+        tmux_env = os.environ.pop('TMUX')
+
+        if has_gte_version('2.6'):
+            set_layout_hook(builder.session, 'client-session-changed')
+
+        builder.session.switch_client()  # switch client to new session
+
+        os.environ['TMUX'] = tmux_env  # set TMUX back again
+    else:
+        if has_gte_version('2.6'):
+            # if attaching for first time
+            set_layout_hook(builder.session, 'client-attached')
+
+            # for cases where user switches client for first time
+            set_layout_hook(builder.session, 'client-session-changed')
+
+        if not detached:
+            builder.session.attach_session()
+
+
+def _load_detached(builder):
+    """
+    Load config in new session but don't attach
+
+    Parameters
+    ----------
+    builder: :class:`workspacebuilder.WorkspaceBuilder`
+    """
+    builder.build()
+
+    if has_gte_version('2.6'):  # prepare for both cases
+        set_layout_hook(builder.session, 'client-attached')
+        set_layout_hook(builder.session, 'client-session-changed')
+
+    print('Session created in detached state.')
+
+
+def _load_append_windows_to_current_session(builder):
+    """
+    Load config as new windows in current session
+
+    Parameters
+    ----------
+    builder: :class:`workspacebuilder.WorkspaceBuilder`
+    """
+    current_attached_session = builder.find_current_attached_session()
+    builder.build(current_attached_session, append=True)
+    if has_gte_version('2.6'):  # prepare for both cases
+        set_layout_hook(builder.session, 'client-attached')
+        set_layout_hook(builder.session, 'client-session-changed')
+
+
+def _setup_plugins(builder):
+    """
+    Runs after before_script
+
+    Parameters
+    ----------
+    builder: :class:`workspacebuilder.WorkspaceBuilder`
+    """
+    for plugin in builder.plugins:
+        plugin.before_script(builder.session)
+
+    return builder.session
 
 
 def load_workspace(
@@ -428,6 +546,7 @@ def load_workspace(
     colors=None,
     detached=False,
     answer_yes=False,
+    append=False,
 ):
     """
     Load a tmux "workspace" session via tmuxp file.
@@ -448,7 +567,11 @@ def load_workspace(
     detached : bool
         Force detached state. default False.
     answer_yes : bool
-        Assume yes when given prompt. default False.
+        Assume yes when given prompt to attach in new session.
+        Default False.
+    append : bool
+       Assume current when given prompt to append windows in same session.
+       Default False.
 
     Notes
     -----
@@ -538,23 +661,26 @@ def load_workspace(
     sconfig = config.trickle(sconfig)
 
     t = Server(  # create tmux server object
-        socket_name=socket_name, socket_path=socket_path,
-        config_file=tmux_config_file, colors=colors,
+        socket_name=socket_name,
+        socket_path=socket_path,
+        config_file=tmux_config_file,
+        colors=colors,
     )
 
     which('tmux')  # raise exception if tmux not found
 
     try:  # load WorkspaceBuilder object for tmuxp config / tmux server
-        builder = WorkspaceBuilder(sconf=sconfig, server=t)
+        builder = WorkspaceBuilder(
+            sconf=sconfig, plugins=load_plugins(sconfig), server=t
+        )
     except exc.EmptyConfigException:
         tmuxp_echo('%s is empty or parsed no config data' % config_file, err=True)
         return
 
     session_name = sconfig['session_name']
 
-    # if the session already exists, prompt the user to attach. tmuxp doesn't
-    # support incremental session building or appending (yet, PR's welcome!)
-    if builder.session_exists(session_name):
+    # if the session already exists, prompt the user to attach
+    if builder.session_exists(session_name) and not append:
         if not detached and (
             answer_yes
             or click.confirm(
@@ -563,42 +689,43 @@ def load_workspace(
                 default=True,
             )
         ):
-            _reattach(builder.session)
+            _reattach(builder)
         return
 
     try:
-        builder.build()  # load tmux session via workspace builder
+        if detached:
+            _load_detached(builder)
+            return _setup_plugins(builder)
+
+        if append:
+            if 'TMUX' in os.environ:  # tmuxp ran from inside tmux
+                _load_append_windows_to_current_session(builder)
+            else:
+                _load_attached(builder, detached)
+
+            return _setup_plugins(builder)
+
+        # append and answer_yes have no meaning if specified together
+        elif answer_yes:
+            _load_attached(builder, detached)
+            return _setup_plugins(builder)
 
         if 'TMUX' in os.environ:  # tmuxp ran from inside tmux
-            if not detached and (
-                answer_yes or click.confirm('Already inside TMUX, switch to session?')
-            ):
-                # unset TMUX, save it, e.g. '/tmp/tmux-1000/default,30668,0'
-                tmux_env = os.environ.pop('TMUX')
+            msg = (
+                "Already inside TMUX, switch to session? yes/no\n"
+                "Or (a)ppend windows in the current active session?\n[y/n/a]"
+            )
+            options = ['y', 'n', 'a']
+            choice = click.prompt(msg, value_proc=_validate_choices(options))
 
-                if has_gte_version('2.6'):
-                    set_layout_hook(builder.session, 'client-session-changed')
-
-                builder.session.switch_client()  # switch client to new session
-
-                os.environ['TMUX'] = tmux_env  # set TMUX back again
-                return builder.session
-            else:  # session created in the background, from within tmux
-                if has_gte_version('2.6'):  # prepare for both cases
-                    set_layout_hook(builder.session, 'client-attached')
-                    set_layout_hook(builder.session, 'client-session-changed')
-
-                sys.exit('Session created in detached state.')
-        else:  # tmuxp ran from inside tmux
-            if has_gte_version('2.6'):
-                # if attaching for first time
-                set_layout_hook(builder.session, 'client-attached')
-
-                # for cases where user switches client for first time
-                set_layout_hook(builder.session, 'client-session-changed')
-
-            if not detached:
-                builder.session.attach_session()
+            if choice == 'y':
+                _load_attached(builder, detached)
+            elif choice == 'a':
+                _load_append_windows_to_current_session(builder)
+            else:
+                _load_detached(builder)
+        else:
+            _load_attached(builder, detached)
 
     except exc.TmuxpException as e:
         import traceback
@@ -616,17 +743,14 @@ def load_workspace(
             builder.session.kill_session()
             tmuxp_echo('Session killed.')
         elif choice == 'a':
-            if 'TMUX' in os.environ:
-                builder.session.switch_client()
-            else:
-                builder.session.attach_session()
+            _reattach(builder)
         else:
             sys.exit()
 
-    return builder.session
+    return _setup_plugins(builder)
 
 
-@click.group(context_settings={'obj': {}})
+@click.group(context_settings={'obj': {}, 'help_option_names': ['-h', '--help']})
 @click.version_option(__version__, '-V', '--version', message='%(prog)s %(version)s')
 @click.option(
     '--log-level',
@@ -880,6 +1004,12 @@ def command_freeze(session_name, socket_name, socket_path, force):
 @click.option(
     '-d', 'detached', help='Load the session without attaching it', is_flag=True
 )
+@click.option(
+    '-a',
+    'append',
+    help='Load configuration, appending windows to the current session',
+    is_flag=True,
+)
 @click.option(
     'colors',
     '-2',
@@ -903,6 +1033,7 @@ def command_load(
     new_session_name,
     answer_yes,
     detached,
+    append,
     colors,
     log_file,
 ):
@@ -942,6 +1073,7 @@ def command_load(
         'answer_yes': answer_yes,
         'colors': colors,
         'detached': detached,
+        'append': append,
     }
 
     if not config:
@@ -949,7 +1081,7 @@ def command_load(
         tmuxp_echo(ctx.get_help(), color=ctx.color)
         ctx.exit()
 
-    if isinstance(config, string_types):
+    if isinstance(config, str):
         load_workspace(config, **tmux_options)
 
     elif isinstance(config, tuple):
@@ -1078,6 +1210,15 @@ def command_convert(confirmed, config):
         print('New config saved to <%s>.' % newfile)
 
 
+@cli.command(name='edit', short_help='Run $EDITOR on config.')
+@click.argument('config', type=ConfigPath(exists=True), nargs=1)
+def command_edit_config(config):
+    config = scan_config(config)
+
+    sys_editor = os.environ.get('EDITOR', 'vim')
+    call([sys_editor, config])
+
+
 @cli.command(
     name='ls', short_help='List configured sessions in {}.'.format(get_config_dir())
 )
